/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-util.c: GUI Utility functions
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include "gui.h"
#include "gui-prefs.h"
#include "gui-install.h"
#include "gui-priority.h"
#include "mirror.h"
#include "subscription.h"

#include "gui-util.h"

#include "libredcarpet.h"
#include "rc-dep-or.h"

GtkWidget *util_create_filebrowser (void);
gboolean util_filebrowser_is_dir (GtkFileSelection * filesel);

typedef struct _RCErrorData RCErrorData;

struct _RCErrorData {
    GtkHTML *html;
    RCError error;
    gboolean returncode;
};

/* FIXME: This should not be a #define */
#define DEFAULT_HOST_ROOT "http://red-carpet.ximian.com"

const char *
rc_gui_get_host (void)
{
    if (appinfo->priority) {
        return rc_gui_priority_host();
    } else {
        char *host = getenv("RC_MAGIC");

        if (host)
            return host;
        else if (appinfo->rcd_host_url)
            return appinfo->rcd_host_url;
        else
            return DEFAULT_HOST_ROOT;
    }
} /* rc_gui_get_host */

void
rc_gui_query_system_packages (void)
{
    RCPackmanError err;
    RCPackageSList *system_packages;

    do {

        rc_debug(RC_DEBUG_LEVEL_INFO, "Querying system packages...\n");
        system_packages = rc_packman_query_all (appinfo->pman);
        rc_debug(RC_DEBUG_LEVEL_INFO, "Querying done!\n");
        err = rc_packman_get_error (appinfo->pman);

        if (err) {
            gchar *reason;
            gchar *freason;

            freason = rc_gui_filter_packman_error (
                rc_packman_get_reason (appinfo->pman));

            reason = g_strdup_printf (
                _("Red Carpet encountered an error querying your system:<br>"
                  "%s"), freason);
            rc_gui_error_box (
                reason, err == RC_PACKMAN_ERROR_FATAL ? RC_ERROR_FATAL :
                RC_ERROR_RETRY);

            g_free (reason);
            g_free (freason);

            if (err == RC_PACKMAN_ERROR_FATAL) {
                gtk_exit (1);
            }
        }
    } while (err);

/* This #if 0 is because of the new RCWorld stuff. */
#if 0
    rc_package_slist_free (appinfo->system_packages);
#else
    g_slist_free (appinfo->system_packages);
#endif
    appinfo->system_packages = system_packages;

    rc_dep_or_process_system_and_channels(appinfo->system_packages,
                                          appinfo->channel_list);
}

void
rc_gui_update_world (void)
{
    RCWorld *world = rc_get_world ();
    
    rc_world_remove_packages (world, RC_WORLD_SYSTEM_PACKAGES);
    rc_world_add_packages_from_slist (world, appinfo->system_packages);

#if 0
    {
        static int count = 1;
        char *filename = g_strdup_printf ("/tmp/rcworld.%d", count);
        FILE *out;

        ++count;
        g_message ("queried system packages, world state in %s", filename);

        out = fopen (filename, "w");
        rc_world_spew (rc_get_world (), out);
        fclose (out);

        g_free (filename);
    }
#endif
}


void
rc_gui_set_page(char *notebook, int page)
{
    gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET(notebook)), page);
} /* rc_gui_set_page */

static char *
format_size(unsigned int size, gboolean nbsp)
{
    char *output;

    if (size > 1024*1024) {
        output = g_strdup_printf(
            _("%.2f%sMB"), (float) size / ((float) 1024*1024),
            nbsp ? "&nbsp;" : " ");
    }
    else if (size > 1024) {
        output = g_strdup_printf(
            _("%d%sKB"), size / 1024, nbsp ? "&nbsp;" : " ");
    }
    else
        output = g_strdup_printf(_("%d%sbytes"), size, nbsp ? "&nbsp;" : " ");

    return output;
} /* format_size */

char *
rc_gui_format_size(unsigned int size)
{
    char *output;

    output = format_size(size, FALSE);

    return output;
} /* rc_gui_format_size */

char *
rc_gui_format_size_with_nbsp(unsigned int size)
{
    char *output;

    output = format_size(size, TRUE);

    return output;
} /* rc_gui_format_size_with_nbsp */

char *
rc_gui_number_to_word(int number)
{
    char *ntable[] = {
        _("zero"),
        _("one"),
        _("two"),
        _("three"),
        _("four"),
        _("five"),
        _("six"),
        _("seven"),
        _("eight"),
        _("nine")
    };
    char *n;

    if (number > 9 || number < 0) {
        n = g_strdup_printf("%d", number);
        return n;
    }

    n = g_strdup(ntable[number]);
    return n;
} /* rc_gui_number_to_word */

const gchar *
rc_gui_section_icon_name (RCPackageSection section)
{
    static const gchar *section_names[RC_SECTION_LAST] = {
        "file://"SHAREDIR"/sections/section-office.png",
        "file://"SHAREDIR"/sections/section-imaging.png",
        "file://"SHAREDIR"/sections/section-pim.png",
        "file://"SHAREDIR"/sections/section-game.png",
        "file://"SHAREDIR"/sections/section-multimedia.png",
        "file://"SHAREDIR"/sections/section-internet.png",
        "file://"SHAREDIR"/sections/section-util.png",
        "file://"SHAREDIR"/sections/section-system.png",
        "file://"SHAREDIR"/sections/section-doc.png",
        "file://"SHAREDIR"/sections/section-devel.png",
        "file://"SHAREDIR"/sections/section-develutil.png",
        "file://"SHAREDIR"/sections/section-library.png",
        "file://"SHAREDIR"/sections/section-xapp.png",
        "file://"SHAREDIR"/sections/section-misc.png"
    };

    return (section_names[section]);
}

const char *
rc_gui_get_pretty_name(const char *package_name)
{
    gboolean pretty = gui_prefs_get_pretty_names();
    const char *name;

    g_return_val_if_fail(package_name, NULL);

    if (pretty)
        name = rc_pretty_name_lookup(package_name);
    else
        name = package_name;

    return name;
} /* rc_gui_get_package_name */

/* CB */
static void
error_ok_clicked(GtkWidget *button, RCErrorData *errordata)
{
    int page = GPOINTER_TO_INT(gtk_object_get_data(
        GTK_OBJECT(errordata->html), "prev_page"));

    if (errordata->error == RC_ERROR_FATAL) {
        /* FIXME: Perhaps exit with a more meaningful error code? */
        gtk_exit(1);
    }
    else if (errordata->error == RC_ERROR_RETRY) {
        gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
        gtk_object_set_data(
            GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
        errordata->returncode = TRUE;
    }
    else if (errordata->error == RC_ERROR_ABORT) {
        gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
        gtk_object_set_data(
            GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
        errordata->returncode = FALSE;
    }
    gtk_main_quit();
} /* error_ok_clicked */

/* CB */
static void
error_cancel_clicked(GtkWidget *button, RCErrorData *errordata)
{
    int page;

    g_assert(errordata->error == RC_ERROR_RETRY);

    page = GPOINTER_TO_INT(gtk_object_get_data(
        GTK_OBJECT(errordata->html), "prev_page"));
    gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
    gtk_object_set_data(
        GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
    errordata->returncode = FALSE;
    gtk_main_quit();
} /* error_cancel_clicked */

/* Display an HTML error box. The message passed in should be HTML. */
gboolean
rc_gui_error_box(const char *message, RCError err)
{
    GtkWidget *frame;
    GtkWidget *html;
    GtkWidget *notebook;
    GtkHTMLStream *stream;
    int page;
    gchar *hstr;
    RCErrorData errordata;
    gboolean startup, mainw;

    startup = GTK_WIDGET_VISIBLE(GET_WIDGET("startup_window"));
    mainw = GTK_WIDGET_VISIBLE(GET_WIDGET("main_window"));

    if (startup)
        gtk_widget_hide(GET_WIDGET("startup_window"));

    if (!mainw)
        gtk_widget_show(GET_WIDGET("main_window"));

    notebook = GET_WIDGET("main_notebook");

    page = gtk_notebook_current_page(GTK_NOTEBOOK(notebook));
    rc_gui_set_page ("main_notebook", PAGE_ERROR);

    gtk_widget_hide(GET_WIDGET("cancel_dialog"));

    frame = GET_WIDGET("error_html");
    g_assert(frame);
    html = gtk_object_get_user_data(GTK_OBJECT(frame));
    g_assert(html);
    gtk_html_allow_selection(GTK_HTML(html), TRUE);

    gtk_object_set_data(GTK_OBJECT(html), "prev_page", GINT_TO_POINTER(page));

    stream = gtk_html_begin(GTK_HTML(html));
    hstr = g_strdup_printf(
        "<html><body marginwidth=0 marginheight=0>"
        "<table cellspacing=0 cellpadding=1 border=0 bgcolor=\"#ffffff\""
        "    width=\"100%%\">"
        "<tr bgcolor=\"#000000\"><td align=left width=36>"
        "<img width=36 height=36 src=\"file://" SHAREDIR "/warning.png\"></td>"
        "<td width=\"100%%\" align=left>"
        "<font color=\"#ffffff\" size=+2><b>%s</b></font>"
        "</td></tr>"
        "</table>"
        "<font color=\"#000000\">"
        "<table cellspacing=0 cellpadding=1 border=0 bgcolor=\"#ffffff\""
        "    width=\"100%%\">"
        "<tr bgcolor=\"#e5e3e3\"><td colspan=2>"
        "<p>%s</p>"             /* "Red Carpet has encountered ..." */
        "<p>%s</p>"
        "</td></tr>"
        "<tr background=\"file://"SHAREDIR"/warning-bar.png\">"
        "<td colspan=2 height=15>&nbsp;"
        "</td></tr>"
        "<tr><td colspan=2><p>%s<ul>", /* Possible actions... */
        err == RC_ERROR_FATAL ? _("Fatal Error") : _("Error"),
        _("Red Carpet has encountered an error while trying to process your "
          "request:"),
        message,
        _("Possible actions:"));
    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));
    g_free (hstr);

    if (err == RC_ERROR_FATAL) {
        hstr = _("<li>Click <b>Quit</b> to quit.");
        gtk_widget_hide(GET_WIDGET("error_cancel"));
        gtk_label_set_text(
            GTK_LABEL(GTK_BIN(GET_WIDGET("error_ok"))->child), _("Quit"));
    }
    else if (err == RC_ERROR_RETRY) {
        hstr = g_strconcat (
            _("<li>Click <b>Retry</b> to retry."),
            _("<li>Click <b>Cancel</b> to cancel."),
            NULL);
        gtk_widget_show(GET_WIDGET("error_ok"));
        gtk_label_set_text(
            GTK_LABEL(GTK_BIN(GET_WIDGET("error_ok"))->child), _("Retry"));
    }
    else if (err == RC_ERROR_ABORT) {
        hstr = _("<li>Click <b>Cancel</b> to cancel.");
        /* uhhhhh -- don't ask */
        gtk_label_set_text(
            GTK_LABEL(GTK_BIN(GET_WIDGET("error_ok"))->child), _("Cancel"));
        gtk_widget_hide(GET_WIDGET("error_cancel"));
    }

    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));

    hstr = "</ul></td></tr></table></font></body></html>";
    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));

    gtk_html_end(GTK_HTML(html), stream, GTK_HTML_STREAM_OK);

    errordata.html = GTK_HTML(html);
    errordata.error = err;
    /* This will determine how to return from this function and is set in the
       callback. We are just initializing it here. */
    errordata.returncode = FALSE;

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("error_ok")), "clicked",
        GTK_SIGNAL_FUNC(error_ok_clicked), &errordata);
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("error_cancel")), "clicked",
        GTK_SIGNAL_FUNC(error_cancel_clicked), &errordata);

    gtk_main();

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("error_ok")), 
        GTK_SIGNAL_FUNC(error_ok_clicked), &errordata);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("error_cancel")), 
        GTK_SIGNAL_FUNC(error_cancel_clicked), &errordata);

    if (startup) {
        gtk_widget_show_now(GET_WIDGET("startup_window"));
        gtk_widget_draw(GET_WIDGET("startup_window"), NULL);
    }

    if (!mainw)
        gtk_widget_hide(GET_WIDGET("main_window"));
    
    return errordata.returncode;
} /* rc_gui_error_box */

void
rc_gui_split_channel_list (GUIChannelSList  *channels,
                           RCChannelSList  **subscribed_channels,
                           RCChannelSList  **unsubscribed_channels)
{
    GUIChannelSList *citer;

    for (citer = channels; citer; citer = citer->next) {
        GUIChannel *channel = (GUIChannel *)(citer->data);

        if (channel->subscribed) {
            *subscribed_channels = g_slist_append (
                *subscribed_channels, RCCHANNEL (channel));
        } else {
            *unsubscribed_channels = g_slist_append (
                *unsubscribed_channels, RCCHANNEL (channel));
        }
    }
}

int
rc_gui_split_by_package_category(RCPackageSList *packages, 
                                 RCPackageSList ***split_packages)
{
    GSList *p;
    int i;
    int count = 0;

    g_return_val_if_fail(split_packages, 0);

    /* RC_SECTION_LAST is just an index for the number of categories we have */
    *split_packages = g_new0(RCPackageSList *, RC_SECTION_LAST);
    p = packages;
    while (p) {
        RCPackage *package = p->data;

        (*split_packages)[package->section] = g_slist_prepend(
            (*split_packages)[package->section], package);
        count++;
        
        p = p->next;
    }
    
    /* Walk through and reverse the lists... */
    for (i = 0; i < RC_SECTION_LAST; i++)
        (*split_packages)[i] = g_slist_reverse((*split_packages)[i]);

    return count;
} /* rc_gui_split_by_package_category */

/* FIXME: this came from rc-package.c, but it doesn't fit well there,
 * because it needs to have a reference to a packman object to compare
 * RCPackageSpec's well... this probably belongs somewhere more
 * general in the end, but it had to come out of rc-package.c ... */

static RCPackageImportance
rc_package_get_highest_importance (RCPackage *package)
{
    RCPackage *system_package;
    RCPackageUpdateSList *iter;
    RCPackageImportance highest = RC_IMPORTANCE_INVALID;

    system_package = rc_package_new ();

    system_package->spec.name = g_strdup (package->spec.name);

    system_package = rc_packman_query (appinfo->pman, system_package);

    for (iter = package->history; iter; iter = iter->next) {
        RCPackageUpdate *update = (RCPackageUpdate *)(iter->data);

        if (rc_packman_version_compare (
                appinfo->pman, RC_PACKAGE_SPEC (update),
                RC_PACKAGE_SPEC (system_package)) <= 0) {
            break;
        }

        if (update->importance < highest || highest == RC_IMPORTANCE_INVALID) {
            highest = update->importance;
        }
    }

    /* The XML is broken: the importance is invalid. In this case,
       throw a g_warning and set the importance to RC_IMPORTANCE_MINOR */
    if (highest == RC_IMPORTANCE_INVALID) {
        g_warning(
            "Update \"%s\" has invalid importance. Setting to minor.",
            system_package->spec.name);
        highest = RC_IMPORTANCE_MINOR;
    }

    rc_package_free (system_package);
    return (highest);
}

int
rc_gui_split_by_package_importance(RCPackageSList *available_updates,
                                   RCPackageSList *system_packages,
                                   RCPackageSList ***split_packages,
                                   int *total_size)
{
    GSList *p;
    int i;
    int count = 0;
    int tot = 0;

    g_return_val_if_fail(split_packages, 0);

    *split_packages = g_new0(RCPackageSList *, RC_IMPORTANCE_LAST);
    p = available_updates;
    while (p) {
        RCPackage *package = p->data;
        RCPackageUpdate *update = package->history->data;
        RCPackageImportance high;

        high = rc_package_get_highest_importance(package);

        /* We can add this to a list of available updates */
        (*split_packages)[high] = g_slist_append(
            (*split_packages)[high], package);
        tot += update->package_size;
        count++;
        
        p = p->next;
    }

    /* Walk through and reverse the lists... */
    for (i = 0; i < RC_IMPORTANCE_LAST; i++)
        (*split_packages)[i] = g_slist_reverse((*split_packages)[i]);

    if (total_size)
        *total_size = tot;

    return count;
} /* split_by_package_importance */



void
rc_gui_set_refresh_item(RefreshType type)
{
    GtkWidget *w;

    w = GET_WIDGET("channel_refresh");
    if (type == REFRESH_CHANNEL) {
        gtk_widget_show(w);
        gtk_widget_set_sensitive(w, TRUE);
    }
    else {
        gtk_widget_hide(w);
        gtk_widget_set_sensitive(w, FALSE);
    }

    w = GET_WIDGET("summary_refresh");
    if (type == REFRESH_SUMMARY) {
        gtk_widget_show(w);
        gtk_widget_set_sensitive(w, TRUE);
    }
    else {
        gtk_widget_hide(w);
        gtk_widget_set_sensitive(w, FALSE);
    }

    w = GET_WIDGET("news_refresh");
    if (type == REFRESH_NEWS) {
        gtk_widget_show(w);
        gtk_widget_set_sensitive(w, TRUE);
    }
    else {
        gtk_widget_hide(w);
        gtk_widget_set_sensitive(w, FALSE);
    }
} /* rc_gui_set_refresh_item */
    
void
rc_gui_sensitize_select_items(SelectType st)
{
    SelectType select_type;
    GtkWidget *w;

    if (appinfo->not_root) {
        /* These should always be desensitized when not root. */
        select_type = SELECT_NEITHER;
    }
    else
        select_type = st;

    w = GET_WIDGET("select_all");
    gtk_widget_set_sensitive(w, select_type & SELECT_ITEM_ALL);

    w = GET_WIDGET("select_none");
    gtk_widget_set_sensitive(w, select_type & SELECT_ITEM_NONE);
} /* rc_gui_sensitize_select_items */

static void
install_local_files (GSList *files)
{
    RCPackage *pkg;
    RCPackageUpdate *pkg_up;
    RCPackageSList *pkgs = NULL;
    RCChannel *dummy_channel;
    GUIChannel *dummy_gui_channel;
    gboolean err = FALSE;
    GSList *problem_files = NULL;

    /* Set up some channels */
    dummy_channel = rc_channel_new ();
    dummy_channel->id = (guint32) -1;
    dummy_channel->name = g_strdup ("Local Packages");

    while (files) {
        gchar *file_name = (gchar *) files->data;

        pkg = rc_packman_query_file (appinfo->pman, file_name);
        if (pkg) {
            struct stat pkgstat;
            
            stat(file_name, &pkgstat);
            
            pkg_up = rc_package_update_new ();
            pkg_up->package_url = g_strconcat ("file://", file_name, NULL);
            pkg_up->importance = RC_IMPORTANCE_SUGGESTED;
            pkg_up->description = g_strdup ("No information available.");
            pkg_up->package_size = pkgstat.st_size;
            
            pkg->history = g_slist_append (pkg->history, pkg_up);
            pkg->channel = dummy_channel;
            
            pkgs = g_slist_prepend (pkgs, pkg);
        }
        else {
            err = TRUE;
            problem_files = g_slist_append(problem_files, file_name);
        }
            
        files = files->next;
    }
        
    if (err) {
        GtkWidget *box;
        GSList *i;
        GString *string;
            
        string = g_string_new(
            _("Red Carpet was unable to install the following packages.\n"
              "Please ensure that they exist and are valid package "
              "files.\n\n"));
            
        i = problem_files;
        while (i) {
            string = g_string_append(string, (char *) i->data);
            string = g_string_append_c(string, '\n');
                
            i = i->next;
        }
            
        box = gnome_message_box_new(
            string->str,
            GNOME_MESSAGE_BOX_ERROR,
            GNOME_STOCK_BUTTON_OK,
            NULL);
        gnome_dialog_run(GNOME_DIALOG(box));
        return;
    }
    
    if (!pkgs) {
        return;
    }

    rc_world_add_packages_from_slist (rc_get_world (), pkgs);

    /* To prevent a crash when sending a dependency report.  Note: this makes
     * me feel ill.
     */
    gtk_object_set_data(
        GTK_OBJECT(appinfo->update_html), "local_installs", pkgs);
        
    dummy_gui_channel = rc_gui_channel_new (dummy_channel);
    gui_install_packages (dummy_gui_channel, pkgs);

    rc_gui_show_executive_summary();

    g_free (dummy_gui_channel);

    rc_world_remove_packages (rc_get_world (), dummy_channel);

    rc_package_slist_free (pkgs);
    rc_channel_free (dummy_channel);
}
            
void
rc_gui_install_local_file (void)
{
    GtkWidget *fs;
    char *match;

    fs = util_create_filebrowser ();
    /* FIXME -- this should prepend arch */

#if 0
    /* FIXME: This will leak. */
    match = g_strdup_printf(
        "*.%s", rc_packman_get_file_extension(appinfo->pman));

    gtk_file_selection_complete(GTK_FILE_SELECTION(fs), match);
#endif

    gtk_widget_show(fs);
} /* rc_gui_install_local_file */

void
rc_gui_set_menubar_sensitive (gboolean sense)
{
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("search"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), sense);

    gtk_widget_set_sensitive(GET_WIDGET("verify_installed_packages"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("install_local_file1"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("about1"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("exit1"), TRUE);
}


/*
 * Support bits for install local file;
 *
 * some of these chunks are taken from xmms.  they have to be done this way
 * because there is no decent file selector, and none that supports multiple
 * file selections.
 */

gboolean util_filebrowser_is_dir (GtkFileSelection * filesel)
{
	gchar *text;
	struct stat buf;
	
	text = g_strdup(gtk_file_selection_get_filename(filesel));
	g_strstrip(text);

	if (stat(text, &buf) == 0)
	{
		if (S_ISDIR(buf.st_mode))
		{		/* Selected directory */
			gtk_file_selection_set_filename(filesel, text);
			g_free(text);
			return TRUE;
		}
	}
	g_free(text);
	return FALSE;
}


static void
filebrowser_install_files (GtkFileSelection *filesel)
{
	GList *sel_list = NULL, *node;
	gchar *text, *filesel_path = NULL;
    GSList *files_to_install = NULL;

    if (util_filebrowser_is_dir (filesel))
        return;

	node = GTK_CLIST (filesel->file_list)->selection;
	while (node) {
		sel_list = g_list_prepend (sel_list, node->data);
		node = node->next;
	}

    filesel_path = g_strdup (gtk_file_selection_get_filename (filesel));
    text = strrchr (filesel_path, '/');
    if (text) text[1] = '\0';   /* one past the / */

    if (sel_list)
    {
        node = sel_list;
        while (node)
        {
            gtk_clist_get_text (GTK_CLIST (filesel->file_list),
                                GPOINTER_TO_INT (node->data), 0, &text);
            files_to_install = g_slist_prepend
                (files_to_install,
                 filesel_path ? g_strconcat (filesel_path, text, NULL) : g_strdup (text));
            node = node->next;
        }
    } else {
        text = gtk_file_selection_get_filename (filesel);
        files_to_install = g_slist_prepend (files_to_install, g_strdup (text));
    }

	gtk_widget_destroy (GTK_WIDGET (filesel));

    install_local_files (files_to_install);

    g_slist_foreach (files_to_install, (GFunc) g_free, NULL);
    g_slist_free (files_to_install);
}

static void filebrowser_ok(GtkWidget * w, GtkWidget * filesel)
{
	gtk_widget_hide(filesel);
	filebrowser_install_files (GTK_FILE_SELECTION(filesel));
}

GtkWidget *
util_create_filebrowser(void)
{
	GtkWidget *filebrowser, *label;

	filebrowser = gtk_file_selection_new(_("Load file(s)"));

	gtk_clist_set_selection_mode(GTK_CLIST(GTK_FILE_SELECTION(filebrowser)->file_list),
                                 GTK_SELECTION_EXTENDED);
	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filebrowser)->ok_button),
                       "clicked", GTK_SIGNAL_FUNC(filebrowser_ok), filebrowser);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filebrowser)->cancel_button),
                              "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy),
                              GTK_OBJECT(filebrowser));

	/*
	 * This is ugly.
	 * Change the Cancel buttons caption to Close.
	 */
	
	label = gtk_label_new(_("Close"));
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
	gtk_container_remove(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button),
                         gtk_container_children(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button))->data);
	gtk_container_add(GTK_CONTAINER(GTK_FILE_SELECTION(filebrowser)->cancel_button), label);
	gtk_widget_show(label);

	gtk_widget_show(filebrowser);
	return filebrowser;
}

gchar *
rc_gui_filter_packman_error (const gchar *error)
{
    guint error_copy_length = 0;
    gchar *error_copy, *error_copy_ptr;
    const gchar *tmp, *error_ptr;

    for (tmp = error; *tmp; tmp++) {
        error_copy_length++;
        if (*tmp == '\n') {
            error_copy_length += 3;
        }
    }

    error_copy = g_new (gchar, error_copy_length + 1);
    error_copy[error_copy_length] = '\0';
    error_copy_ptr = error_copy;

    error_ptr = error;

    while ((tmp = strchr (error_ptr, '\n'))) {
        strncpy (error_copy_ptr, error_ptr, tmp - error_ptr);
        error_copy_ptr += (tmp - error_ptr);

        strncpy (error_copy_ptr, "<br>", 4);
        error_copy_ptr += 4;

        error_ptr = tmp + 1;
    }

    if (error_ptr) {
        strcpy (error_copy_ptr, error_ptr);
    }

    return (error_copy);
}

void
rc_gui_restart (gboolean with_dialog)
{
    const char *cmd_name = appinfo->argv[0];
    GtkWidget *dialog;

    if (!cmd_name) {
        g_error(
            "Didn't have argv0, can't restart.  Please restart RC manually!");
    }

    if (with_dialog) {
        dialog = gnome_message_box_new(
            _("Ximian Red Carpet will now restart."),
            GNOME_MESSAGE_BOX_INFO,
            GNOME_STOCK_BUTTON_OK,
            NULL);
        gnome_dialog_run_and_close (GNOME_DIALOG(dialog));
    }

    execvp (cmd_name, appinfo->argv);
} /* rc_gui_restart */
